# Copyright (c) 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from neutron.plugins.ml2 import driver_context
from neutron.tests.unit.plugins.ml2 import test_plugin

from networking_ansible import config
from networking_ansible.ml2 import mech_driver

QUOTA_REGISTRIES = (
    "neutron.quota.resource_registry.unregister_all_resources",
    "neutron.quota.resource_registry.register_resource_by_name",
)


class NetworkingAnsibleTestCase(test_plugin.Ml2PluginV2TestCase):
    def setUp(self):
        # This is to avoid "No resource found" messages printed to stderr from
        # quotas Neutron code.
        for func in QUOTA_REGISTRIES:
            mock.patch(func).start()
        super(NetworkingAnsibleTestCase, self).setUp()
        self.ansconfig = config
        self.testmac = '01:23:45:67:89:ab'
        self.testhost = 'testhost'
        self.inventory = {
            'all': {
                'hosts': {
                    self.testhost: {
                        'mac': self.testmac
                    }
                }
            }
        }

        with mock.patch('networking_ansible.ml2.mech_driver.config') as m_cfg:
            m_cfg.build_ansible_inventory.return_value = self.inventory
            self.mech = mech_driver.AnsibleMechanismDriver()
            self.mech.initialize()
        self.testsegid = '37'
        self.testsegid2 = '73'
        self.testsegids = [73, 7, 3]
        self.testport = 'switchportid'
        self.empty_inventory = {'all': {'hosts': {}}}

        # Define mocked network context
        self.mock_net_context = mock.create_autospec(
            driver_context.NetworkContext).return_value
        self.mock_net_context.current = {
            'id': 37,
            'provider:network_type': 'vlan',
            'provider:segmentation_id': self.testsegid,
            'provider:physical_network': 'physnet',
        }

        # define mocked port context
        self.mock_port_context = mock.create_autospec(
            driver_context.PortContext).return_value
        self.lli_no_mac = {
            'local_link_information': [{
                'switch_info': self.testhost,
                'port_id': self.testport,
            }]
        }
        self.lli_no_info = {
            'local_link_information': [{
                'switch_id': self.testmac,
                'port_id': self.testport,
            }]
        }
        self.mock_port_context.current = {
            'id': 'aaaa-bbbb-cccc',
            'binding:profile': self.lli_no_mac,
            'binding:vnic_type': 'baremetal',
            'binding:vif_type': 'other',
        }
        self.mock_port_context._plugin_context = mock.MagicMock()
        self.mock_port_context.network = mock.Mock()
        self.mock_port_context.network.current = {
            'id': 'aaaa-bbbb-cccc',
            # TODO(radez) should an int be use here or str ok?
            'provider:segmentation_id': self.testsegid,
        }
        self.mock_port_context.segments_to_bind = [
            self.mock_port_context.network.current
        ]



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_ansible.tests.unit import base


class MockedConfigParser(mock.Mock):

    def __init__(self, conffile, sections):
        super(MockedConfigParser, self).__init__()
        self.conffile = conffile
        self.sections = sections

    @classmethod
    def _parse_file(cls, config_file, namespace):
        return

    def parse(self):
        section_data = {'ansible:testhost': {'mac': ['01:23:45:67:89:ab']}}
        if self.conffile == 'foo2':
            section_data = {
                'ansible:h1': {'manage_vlans': ['0']},
                'ansible:h2': {'manage_vlans': ['true']},
                'ansible:h3': {'manage_vlans': ['false']},
            }

        self.sections.update(section_data)


class TestBuildAnsibleInventory(base.NetworkingAnsibleTestCase):

    def test_build_ansible_inventory_empty_hosts(self):
        self.assertEqual(self.empty_inventory,
                         self.ansconfig.build_ansible_inventory())

    @mock.patch('networking_ansible.config.LOG')
    @mock.patch('networking_ansible.config.cfg.ConfigParser')
    def test_build_ansible_inventory_parser_error(self, mock_parser, mock_log):
        mock_parser().parse.side_effect = IOError()
        self.assertEqual(self.empty_inventory,
                         self.ansconfig.build_ansible_inventory())
        mock_log.error.assert_called()

    @mock.patch('networking_ansible.config.cfg.ConfigParser',
                MockedConfigParser)
    def test_build_ansible_inventory_w_hosts(self):
        self.assertEqual(self.inventory,
                         self.ansconfig.build_ansible_inventory())

    @mock.patch('networking_ansible.config.cfg.ConfigParser',
                MockedConfigParser)
    @mock.patch('neutron.common.test_lib.test_config')
    def test_build_ansible_inventory_from_file(self, test_config):
        test_config.get.return_value = ['foo2']
        self.setup_config()

        ansible_inventory = self.ansconfig.build_ansible_inventory()

        hosts = ansible_inventory['all']['hosts']
        self.assertEqual({'manage_vlans': False}, hosts['h1'])
        self.assertEqual({'manage_vlans': True}, hosts['h2'])
        self.assertEqual({'manage_vlans': False}, hosts['h3'])



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_ansible import exceptions
from networking_ansible.tests.unit import base


class TestCreateDeleteNetwork(base.NetworkingAnsibleTestCase):

    @mock.patch('networking_ansible.ansible_networking'
                '.AnsibleNetworking._run_task')
    def test_create_network(self, mock_run_task):
        self.mech.ansnet.create_network(self.testhost, self.testsegid)
        mock_run_task.assert_called_once_with('create_network',
                                              self.testhost,
                                              self.testsegid)

    @mock.patch('networking_ansible.ansible_networking'
                '.AnsibleNetworking._run_task')
    def test_delete_network(self, mock_run_task):
        self.mech.ansnet.delete_network(self.testhost, self.testsegid)
        mock_run_task.assert_called_once_with('delete_network',
                                              self.testhost,
                                              self.testsegid)


@mock.patch('networking_ansible.ansible_networking.ansible_runner')
class TestRunTask(base.NetworkingAnsibleTestCase):
    def test_run_task_no_switchport(self, mock_ans_runner):
        mock_result = mock_ans_runner.run.return_value
        mock_result.stats = {'failures': []}

        self.mech.ansnet._run_task('fake_task',
                                   self.testhost,
                                   self.testsegid)
        # Assert switch_port is not set
        self.assertNotIn('port_name',
                         mock_ans_runner.run.call_args[1]['playbook'][0]
                                                      ['tasks'][0]['vars']
                         )
        # Assert switch_port is not set
        self.assertNotIn('port_description',
                         mock_ans_runner.run.call_args[1]['playbook'][0]
                                                      ['tasks'][0]['vars']
                         )

    def test_run_task_w_switchport(self, mock_ans_runner):
        mock_result = mock_ans_runner.run.return_value
        mock_result.stats = {'failures': []}

        self.mech.ansnet._run_task('fake_task',
                                   self.testhost,
                                   self.testsegid,
                                   'fake_switchport')
        # Assert switch_port is set
        self.assertEqual(
            'fake_switchport',
            mock_ans_runner.run.call_args[1]['playbook'][0]['tasks']
                                         [0]['vars']['port_name'])
        # Assert switch_port is set
        self.assertEqual(
            'fake_switchport',
            mock_ans_runner.run.call_args[1]['playbook'][0]['tasks']
                                         [0]['vars']['port_description'])

    def test_run_task_w_segmentation_id_1(self, mock_ans_runner):
        mock_result = mock_ans_runner.run.return_value
        mock_result.stats = {'failures': []}

        self.mech.ansnet._run_task('fake_task',
                                   self.testhost,
                                   '1',
                                   'fake_switchport')
        # Assert seg name is default
        self.assertEqual(
            'default',
            mock_ans_runner.run.call_args[1]['playbook'][0]['tasks']
                                         [0]['vars']['segmentation_name'])

    def test_run_task_failures(self, mock_ans_runner):
        mock_result = mock_ans_runner.run.return_value
        mock_result.status = 'failed'
        self.assertRaises(exceptions.AnsibleRunnerException,
                          self.mech.ansnet._run_task,
                          'fake_task',
                          self.testhost,
                          self.testsegid,
                          'fake_switchport')

        mock_result.status = ''
        mock_result.stats = {'failures': ['I got some failure']}
        self.assertRaises(exceptions.AnsibleRunnerException,
                          self.mech.ansnet._run_task,
                          'fake_task',
                          self.testhost,
                          self.testsegid,
                          'fake_switchport')


@mock.patch('networking_ansible.ansible_networking'
            '.AnsibleNetworking._run_task')
class TestVlanAccessPort(base.NetworkingAnsibleTestCase):

    def test_assign_vlan_access_port(self, mock_run_task):
        self.mech.ansnet.vlan_access_port('assign',
                                          self.mock_port_context.current,
                                          self.mock_net_context.current)
        mock_run_task.assert_called_once_with('update_port',
                                              self.testhost,
                                              self.testsegid,
                                              self.testport)

    def test_remove_vlan_access_port(self, mock_run_task):
        self.mech.ansnet.vlan_access_port('remove',
                                          self.mock_port_context.current,
                                          self.mock_net_context.current)
        mock_run_task.assert_called_once_with('delete_port',
                                              self.testhost,
                                              self.testsegid,
                                              self.testport)

    def test_remove_vlan_access_port_wo_link_local(self, mock_run_task):
        port = self.mock_port_context.current
        del port['binding:profile']['local_link_information']
        self.mech.ansnet.vlan_access_port('remove',
                                          self.mock_port_context.current,
                                          self.mock_net_context.current)
        mock_run_task.assert_not_called()

    def test_remove_vlan_access_port_raises(self, mock_run_task):
        mock_run_task.side_effect = exceptions.AnsibleRunnerException('test')
        self.assertRaises(exceptions.AnsibleRunnerException,
                          self.mech.ansnet.vlan_access_port,
                          'remove',
                          self.mock_port_context.current,
                          self.mock_net_context.current)


@mock.patch('networking_ansible.ansible_networking'
            '.AnsibleNetworking._run_task')
class TestConfTrunkPort(base.NetworkingAnsibleTestCase):

    def test_assign_trunk_port(self, m_run_task):
        self.mech.ansnet.conf_trunk_port(self.testhost,
                                         self.testport,
                                         self.testsegid,
                                         trunked_vlans=self.testsegids)

        m_run_task.assert_called_once_with('conf_trunk_port',
                                           self.testhost,
                                           self.testsegid,
                                           switch_port=self.testport,
                                           trunked_vlans=self.testsegids)

    def test_remove_trunk_port(self, m_run_task):
        self.mech.ansnet.vlan_access_port('remove',
                                          self.mock_port_context.current,
                                          self.mock_net_context.current)

        m_run_task.assert_called_once_with('delete_port',
                                           self.testhost,
                                           self.testsegid,
                                           self.testport)

    def test_remove_trunk_port_raises(self, m_run_task):
        m_run_task.side_effect = exceptions.AnsibleRunnerException('test')
        self.assertRaises(exceptions.AnsibleRunnerException,
                          self.mech.ansnet.vlan_access_port, 'remove',
                          self.mock_port_context.current,
                          self.mock_net_context.current)



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import provisioning_blocks
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import trunk_details
from neutron_lib.callbacks import resources
from neutron_lib.plugins.ml2 import api
from oslo_log import log as logging

from networking_ansible import ansible_networking
from networking_ansible import config
from networking_ansible import exceptions

LOG = logging.getLogger(__name__)

ANSIBLE_NETWORKING_ENTITY = 'ANSIBLENETWORKING'


class AnsibleMechanismDriver(api.MechanismDriver):
    """ML2 Mechanism Driver for Ansible Networking

    https://www.ansible.com/integrations/networks
    """

    def initialize(self):
        LOG.debug("Initializing Ansible ML2 driver")

        inventory = config.build_ansible_inventory()
        self.ansnet = ansible_networking.AnsibleNetworking(inventory)

    def create_network_postcommit(self, context):
        """Create a network.

        :param context: NetworkContext instance describing the new
        network.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        """

        network = context.current
        network_id = network['id']
        provider_type = network['provider:network_type']
        segmentation_id = network['provider:segmentation_id']
        # physnet = network['provider:physical_network']

        if provider_type == 'vlan' and segmentation_id:
            # assuming all hosts
            # TODO(radez): can we filter by physnets?
            for host_name in self.ansnet.inventory['all']['hosts']:
                host = self.ansnet.inventory['all']['hosts'][host_name]
                if host.get('manage_vlans', True):
                    # Create VLAN on the switch
                    try:
                        self.ansnet.create_network(host_name, segmentation_id)
                        LOG.info('Network {net_id} has been added on ansible '
                                 'host {host}'.format(
                                     net_id=network['id'],
                                     host=host_name))

                    except Exception as e:
                        # TODO(radez) I don't think there is a message returned
                        #             from ansible runner's exceptions
                        LOG.error('Failed to create network {net_id} '
                                  'on ansible host: {host}, '
                                  'reason: {err}'.format(
                                      net_id=network_id,
                                      host=host_name,
                                      err=e))
                        raise ml2_exc.MechanismDriverError(e)

    def delete_network_postcommit(self, context):
        """Delete a network.

        :param context: NetworkContext instance describing the current
        state of the network, prior to the call to delete it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """

        network = context.current
        provider_type = network['provider:network_type']
        segmentation_id = network['provider:segmentation_id']
        # physnet = network['provider:physical_network']

        if provider_type == 'vlan' and segmentation_id:
            # assuming all hosts
            # TODO(radez): can we filter by physnets?
            for host_name in self.ansnet.inventory['all']['hosts']:
                host = self.ansnet.inventory['all']['hosts'][host_name]
                if host.get('manage_vlans', True):
                    # Delete VLAN on the switch
                    try:
                        self.ansnet.delete_network(host_name, segmentation_id)
                        LOG.info('Network {net_id} has been deleted on '
                                 'ansible host {host}'.format(
                                     net_id=network['id'],
                                     host=host_name))

                    except Exception as e:
                        LOG.error('Failed to delete network {net_id} '
                                  'on ansible host: {host}, '
                                  'reason: {err}'.format(net_id=network['id'],
                                                         host=host_name,
                                                         err=e))

    def update_port_postcommit(self, context):
        """Update a port.

        :param context: PortContext instance describing the new
        state of the port, as well as the original state prior
        to the update_port call.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Raising an exception will
        result in the deletion of the resource.
        update_port_postcommit is called for all changes to the port
        state. It is up to the mechanism driver to ignore state or
        state changes that it does not know or care about.
        """
        port = context.network.current
        if self._is_port_bound(context.current):
            provisioning_blocks.provisioning_complete(
                context._plugin_context, port['id'], resources.PORT,
                ANSIBLE_NETWORKING_ENTITY)
        elif self._is_port_bound(context.original):
            # The port has been unbound. This will cause the local link
            # information to be lost, so remove the port from the network on
            # the switch now while we have the required information.
            self.ansnet.vlan_access_port('remove', context.original, port)

    def delete_port_postcommit(self, context):
        """Delete a port.

        :param context: PortContext instance describing the current
        state of the port, prior to the call to delete it.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        if self._is_port_bound(context.current):
            self.ansnet.vlan_access_port('remove',
                                         context.current,
                                         context.network.current)

    def bind_port(self, context):
        """Attempt to bind a port.

        :param context: PortContext instance describing the port

        This method is called outside any transaction to attempt to
        establish a port binding using this mechanism driver. Bindings
        may be created at each of multiple levels of a hierarchical
        network, and are established from the top level downward. At
        each level, the mechanism driver determines whether it can
        bind to any of the network segments in the
        context.segments_to_bind property, based on the value of the
        context.host property, any relevant port or network
        attributes, and its own knowledge of the network topology. At
        the top level, context.segments_to_bind contains the static
        segments of the port's network. At each lower level of
        binding, it contains static or dynamic segments supplied by
        the driver that bound at the level above. If the driver is
        able to complete the binding of the port to any segment in
        context.segments_to_bind, it must call context.set_binding
        with the binding details. If it can partially bind the port,
        it must call context.continue_binding with the network
        segments to be used to bind at the next lower level.

        If the binding results are committed after bind_port returns,
        they will be seen by all mechanism drivers as
        update_port_precommit and update_port_postcommit calls. But if
        some other thread or process concurrently binds or updates the
        port, these binding results will not be committed, and
        update_port_precommit and update_port_postcommit will not be
        called on the mechanism drivers with these results. Because
        binding results can be discarded rather than committed,
        drivers should avoid making persistent state changes in
        bind_port, or else must ensure that such state changes are
        eventually cleaned up.

        Implementing this method explicitly declares the mechanism
        driver as having the intention to bind ports. This is inspected
        by the QoS service to identify the available QoS rules you
        can use with ports.
        """
        port = context.current
        binding_profile = port['binding:profile']
        local_link_info = binding_profile.get('local_link_information')
        # Validate port and local link info
        if not (self._is_port_supported(port) and local_link_info):
            # TODO(radez) log debug messages here
            return

        segments = context.segments_to_bind

        switch_mac = local_link_info[0].get('switch_id', '').upper()
        switch_name = local_link_info[0].get('switch_info')
        switch_port = local_link_info[0].get('port_id')
        # fill in the switch name if mac exists but name is not defined
        # this provides support for introspection when the switch's mac is
        # also provided in the ML2 conf for ansible-networking
        if not switch_name and switch_mac in self.ansnet.mac_map:
            switch_name = self.ansnet.mac_map[switch_mac]

        network = context.network.current
        # If segmentation ID is None, vlan will be set to 1
        segmentation_id = network['provider:segmentation_id'] or '1'
        segments = context.segments_to_bind

        LOG.debug('Plugging in port {switch_port} on '
                  '{switch_name} to vlan: {segmentation_id}'.format(
                      switch_port=switch_port,
                      switch_name=switch_name,
                      segmentation_id=segmentation_id))

        provisioning_blocks.add_provisioning_component(
            context._plugin_context, port['id'], resources.PORT,
            ANSIBLE_NETWORKING_ENTITY)

        LOG.debug("Putting port {port} on {switch_name} to vlan: "
                  "{segmentation_id}".format(
                      port=switch_port,
                      switch_name=switch_name,
                      segmentation_id=segmentation_id))

        # Assign port to network
        try:
            if trunk_details.TRUNK_DETAILS in port:
                sub_ports = port[trunk_details.TRUNK_DETAILS]['sub_ports']
                trunked_vlans = [sp['segmentation_id'] for sp in sub_ports]
                self.ansnet.conf_trunk_port(switch_name, switch_port,
                                            segmentation_id, trunked_vlans)
            else:
                self.ansnet.vlan_access_port('assign', context.current,
                                             context.network.current)
            context.set_binding(segments[0][api.ID],
                                portbindings.VIF_TYPE_OTHER, {})
            LOG.info('Port {neutron_port} has been plugged into '
                     'network {net_id} on device {switch_name}'.format(
                         neutron_port=port['id'],
                         net_id=network['id'],
                         switch_name=switch_name))
        except Exception as e:
            LOG.error('Failed to plug in port {neutron_port} on '
                      'device: {switch_name} from network {net_id} '
                      'reason: {exc}'.format(
                          neutron_port=port['id'],
                          net_id=network['id'],
                          switch_name=switch_name,
                          exc=e))
            raise ml2_exc.MechanismDriverError(e)

    def _link_info_from_port(self, port, network=None):
        network = network or {}
        # Validate port and local link info
        local_link_info = port['binding:profile'].get('local_link_information')
        if not local_link_info:
            msg = 'local_link_information is missing in port {port_id} ' \
                  'binding:profile'.format(port_id=port['id'])
            LOG.debug(msg)
            raise exceptions.LocalLinkInfoMissingException(msg)

    @staticmethod
    def _is_port_supported(port):
        """Return whether a port is supported by this driver.

        :param port: The port to check
        :returns: Whether the port is supported

        Ports supported by this driver have a VNIC type of 'baremetal'.
        """
        vnic_type = port[portbindings.VNIC_TYPE]
        return vnic_type == portbindings.VNIC_BAREMETAL

    @staticmethod
    def _is_port_bound(port):
        """Return whether a port is bound by this driver.

        Ports bound by this driver have their VIF type set to 'other'.

        :param port: The port to check
        :returns: Whether the port is bound
        """
        if not AnsibleMechanismDriver._is_port_supported(port):
            return False

        vif_type = port[portbindings.VIF_TYPE]
        return vif_type == portbindings.VIF_TYPE_OTHER



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ansible_runner
from oslo_log import log as logging

from networking_ansible import exceptions

LOG = logging.getLogger(__name__)
TRUNK_TASK = 'conf_trunk_port'


class AnsibleNetworking(object):
    """Object to invoke ansible_runner to call Ansible Networking

    Hold inventory and provide an interface for calling
    roles in Ansible Networking to manipulate switch configuration
    """

    def __init__(self, inventory):
        self.inventory = inventory
        # create a dict of switches that have macs defined
        # dict uses mac for key and name for value
        hosts = inventory['all']['hosts']
        self.mac_map = {
            h['mac'].upper(): name for name, h in hosts.items() if 'mac' in h
        }

    def _run_task(self, task, host_name, segmentation_id,
                  switch_port=None, trunked_vlans=None):
        """Run a task.

        :param task: name of task in openstack-ml2 ansible role
        :param host_name: name of a host defined in ml2 conf ini files
        :param segmentation_id: vlan id of the network,
                                default VLAN for trunk ports
        :param switch_port: port name on the switch (optional)
        :param trunked_vlans: A list of VLAN's for trunk port configuration

        See etc/ansible/roles/openstack-ml2/README.md for an exmaple playbook
        """
        # TODO(radez): This is hard coded for juniper because
        #       that was the initial vm we got running.
        #       need to find out how to do this cross vendor
        if segmentation_id == '1':
            segmentation_name = 'default'
        else:
            segmentation_name = 'vlan{}'.format(segmentation_id)

        trunked_vlans = trunked_vlans or []

        # build out the ansible playbook
        playbook = [{
            'name': 'Openstack networking-ansible playbook',
            'hosts': host_name,
            'gather_facts': 'no',  # no need to gather facts every run
            'tasks': [{
                'name': 'do {}'.format(task),
                'import_role': {
                    'name': 'openstack-ml2',
                    'tasks_from': task,
                },
                'vars': {
                    'segmentation_name': segmentation_name,
                    'segmentation_id': segmentation_id,
                }
            }]
        }]

        if task == TRUNK_TASK:
            playbook[0]['tasks'][0]['vars']['trunked_vlans'] = trunked_vlans
        if switch_port:
            playbook[0]['tasks'][0]['vars']['port_name'] = switch_port
            playbook[0]['tasks'][0]['vars']['port_description'] = switch_port

        # TODO(radez) should we pass ident?
        result = ansible_runner.run(playbook=playbook,
                                    inventory=self.inventory,
                                    settings={'pexpect_use_poll': False})
        if result.status == 'failed' or \
                (result.stats and result.stats.get('failures', [])):
            raise exceptions.AnsibleRunnerException(' '.join(result.stdout))
        return result

    def create_network(self, host_name, segmentation_id):
        self._run_task('create_network', host_name, segmentation_id)

    def delete_network(self, host_name, segmentation_id):
        self._run_task('delete_network', host_name, segmentation_id)

    def vlan_access_port(self, assign_remove, port, network):
        """Assign an access port to a vlan.

        If the configuration required to unplug the port is not present
        (e.g. local link information), the port will not be unplugged and no
        exception will be raised.

        :param assign_remove: 'assign' or 'remove'
        :param port: The port to unplug
        :param network: The network from which to unplug the port
        """
        task = {'assign': 'update_port',
                'remove': 'delete_port'}
        debug_msg = {'assign': 'Plugging in port {switch_port} on '
                               '{switch_name} to vlan: {segmentation_id}',
                     'remove': 'Unplugging port {switch_port} on '
                               '{switch_name} from vlan: {segmentation_id}'}
        info_msg = {'assign': 'Port {neutron_port} has been plugged into '
                              'network {net_id} on device {switch_name}',
                    'remove': 'Port {neutron_port} has been unplugged from '
                              'network {net_id} on device {switch_name}'}
        error_msg = {'assign': 'Failed to unplug port {neutron_port} on '
                               'device: {switch_name} from network {net_id} '
                               'reason: {exc}',
                     'remove': 'Failed to plug in port {neutron_port} on '
                               'device: {switch_name} from network {net_id} '
                               'reason: {exc}'}

        # If segmentation ID is None, set vlan 1
        segmentation_id = network['provider:segmentation_id'] or '1'

        local_link_info = port['binding:profile'].get('local_link_information')
        if not local_link_info:
            return
        switch_mac = local_link_info[0].get('switch_id', '').upper()
        switch_name = local_link_info[0].get('switch_info')
        switch_port = local_link_info[0].get('port_id')
        # fill in the switch name if mac exists but name is not defined
        # this provides support for introspection when the switch's mac is
        # also provided in the ML2 conf for ansible-networking
        if not switch_name and switch_mac in self.mac_map:
            switch_name = self.mac_map[switch_mac]
        try:
            LOG.debug(debug_msg[assign_remove].format(
                switch_port=switch_port,
                switch_name=switch_name,
                segmentation_id=segmentation_id))
            self._run_task(task[assign_remove], switch_name,
                           segmentation_id, switch_port)
            LOG.info(info_msg[assign_remove].format(
                neutron_port=port['id'],
                net_id=network['id'],
                switch_name=switch_name))
        except Exception as e:
            LOG.error(error_msg[assign_remove].format(
                neutron_port=port['id'],
                net_id=network['id'],
                switch_name=switch_name,
                exc=e))
            raise

    def conf_trunk_port(self, hostname, port, vlan_id, trunked_vlans):
        """Configure trunk port w/ default vlan and optional additional vlans

        :param hostname: The name of the host in Ansible inventory.
        :param port: The port to configure.
        :param vlan_id: the default VLAN ID to assign to the port
                        An empty or None value will default to the
                        target device's default VLAN assignment. This
                        default is assigned in the ansible role.
        :param trunked_vlans: A list of VLAN IDs to add to the port in
                              addition to the default VLAN.
        """
        return self._run_task(TRUNK_TASK,
                              hostname, vlan_id,
                              switch_port=port,
                              trunked_vlans=trunked_vlans)



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock

from neutron.common import test_lib
from neutron.plugins.ml2.common import exceptions as ml2_exc
# from neutron.tests.unit.plugins.ml2 import test_plugin
# from neutron_lib.api.definitions import portbindings
# from neutron_lib.api.definitions import provider_net
from neutron_lib.api.definitions import trunk_details
# import webob.exc

from networking_ansible import ansible_networking
# from networking_ansible import exceptions
# from networking_ansible.ml2 import exceptions as netans_ml2exc
from networking_ansible.tests.unit import base


class TestLibTestConfigFixture(fixtures.Fixture):
    def __init__(self):
        self._original_test_config = None

    def _setUp(self):
        self.addCleanup(self._restore)
        self._original_test_config = test_lib.test_config.copy()

    def _restore(self):
        if self._original_test_config is not None:
            test_lib.test_config = self._original_test_config


@mock.patch('networking_ansible.ansible_networking.'
            'AnsibleNetworking.vlan_access_port')
@mock.patch('networking_ansible.ml2.mech_driver.provisioning_blocks',
            autospec=True)
class TestBindPort(base.NetworkingAnsibleTestCase):
    def test_bind_port_info_no_mac(self,
                                   mock_prov_blks,
                                   mock_vlan_access_port):
        self.mech.bind_port(self.mock_port_context)
        mock_vlan_access_port.assert_called_once_with(
            'assign',
            self.mock_port_context.current,
            self.mock_port_context.network.current)

    def test_bind_port_mac_no_info_local_link_info(self,
                                                   mock_prov_blks,
                                                   mock_vlan_access_port):
        bind_prof = 'binding:profile'
        self.mock_port_context.current[bind_prof] = self.lli_no_info
        self.mech.bind_port(self.mock_port_context)
        mock_vlan_access_port.assert_called_once()

    @mock.patch('networking_ansible.ml2.mech_driver.'
                'AnsibleMechanismDriver._is_port_supported')
    def test_bind_port_port_not_supported(self,
                                          mock_prov_blks,
                                          mock_vlan_access_port,
                                          mock_port_supported):
        mock_port_supported.return_value = False
        self.mech.bind_port(self.mock_port_context)
        mock_vlan_access_port.assert_not_called()

    def test_bind_port_no_local_link_info(self,
                                          mock_prov_blks,
                                          mock_vlan_access_port):
        bind_prof = 'binding:profile'
        local_link_info = 'local_link_information'
        del self.mock_port_context.current[bind_prof][local_link_info]
        self.mech.bind_port(self.mock_port_context)
        mock_vlan_access_port.assert_not_called()

    @mock.patch.object(ansible_networking.AnsibleNetworking, 'conf_trunk_port')
    def test_bind_port_trunk_port(self,
                                  mock_conf_trunk_port,
                                  mock_prov_blks,
                                  mock_update_access_port):
        td = {'sub_ports': [{'segmentation_id': self.testsegid2}]}
        self.mock_port_context.current[trunk_details.TRUNK_DETAILS] = td
        self.mech.bind_port(self.mock_port_context)
        mock_conf_trunk_port.assert_called_once_with(self.testhost,
                                                     self.testport,
                                                     self.testsegid,
                                                     [self.testsegid2])


class TestIsPortSupported(base.NetworkingAnsibleTestCase):
    def test_is_port_supported(self):
        self.assertTrue(
            self.mech._is_port_supported(self.mock_port_context.current))

    def test_is_port_supported_not_baremetal(self):
        self.mock_port_context.current['binding:vnic_type'] = 'not-baremetal'
        self.assertFalse(
            self.mech._is_port_supported(self.mock_port_context.current))


@mock.patch('networking_ansible.ml2.mech_driver.'
            'AnsibleMechanismDriver._is_port_supported')
class TestIsPortBound(base.NetworkingAnsibleTestCase):
    def test_is_port_bound(self, mock_port_supported):
        self.assertTrue(
            self.mech._is_port_bound(self.mock_port_context.current))

    def test_is_port_bound_not_other(self, mock_port_supported):
        self.mock_port_context.current['binding:vif_type'] = 'not-other'
        self.assertFalse(
            self.mech._is_port_bound(self.mock_port_context.current))

    def test_is_port_bound_port_not_supported(self, mock_port_supported):
        mock_port_supported.return_value = False
        self.assertFalse(
            self.mech._is_port_bound(self.mock_port_context.current))


@mock.patch('networking_ansible.ansible_networking.'
            'AnsibleNetworking.create_network')
class TestCreateNetworkPostCommit(base.NetworkingAnsibleTestCase):
    def test_create_network_postcommit(self, mock_create_network):
        self.mech.create_network_postcommit(self.mock_net_context)
        mock_create_network.assert_called_once()

    def test_create_network_postcommit_manage_vlans_false(self,
                                                          mock_create_network):
        self.inventory['all']['hosts'][self.testhost]['manage_vlans'] = False
        self.mech.create_network_postcommit(self.mock_net_context)
        mock_create_network.assert_not_called()

    def test_create_network_postcommit_fails(self, mock_create_network):
        mock_create_network.side_effect = Exception()
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.mech.create_network_postcommit,
                          self.mock_net_context)
        mock_create_network.assert_called_once()

    def test_create_network_postcommit_not_vlan(self, mock_create_network):
        self.mock_net_context.current['provider:network_type'] = 'not-vlan'
        self.mech.create_network_postcommit(self.mock_net_context)
        mock_create_network.assert_not_called()

    def test_create_network_postcommit_not_segmentation_id(self,
                                                           mock_create_netwrk):
        self.mock_net_context.current['provider:segmentation_id'] = ''
        self.mech.create_network_postcommit(self.mock_net_context)
        mock_create_netwrk.assert_not_called()


@mock.patch('networking_ansible.ansible_networking.'
            'AnsibleNetworking.delete_network')
class TestDeleteNetworkPostCommit(base.NetworkingAnsibleTestCase):
    def test_delete_network_postcommit(self, mock_delete_network):
        self.mech.delete_network_postcommit(self.mock_net_context)
        mock_delete_network.assert_called_once()

    def test_delete_network_postcommit_manage_vlans_false(self,
                                                          mock_delete_network):
        self.inventory['all']['hosts'][self.testhost]['manage_vlans'] = False
        self.mech.delete_network_postcommit(self.mock_net_context)
        mock_delete_network.assert_not_called()

    def test_delete_network_postcommit_fails(self, mock_delete_network):
        mock_delete_network.side_effect = Exception()
        self.mech.delete_network_postcommit(self.mock_net_context)
        mock_delete_network.assert_called_once()

    def test_delete_network_postcommit_not_vlan(self, mock_delete_network):
        self.mock_net_context.current['provider:network_type'] = 'not-vlan'
        self.mech.delete_network_postcommit(self.mock_net_context)
        mock_delete_network.assert_not_called()

    def test_delete_network_postcommit_not_segmentation_id(self,
                                                           mock_delete_netwrk):
        self.mock_net_context.current['provider:segmentation_id'] = ''
        self.mech.delete_network_postcommit(self.mock_net_context)
        mock_delete_netwrk.assert_not_called()


@mock.patch('networking_ansible.ml2.mech_driver.'
            'AnsibleMechanismDriver._is_port_bound')
@mock.patch('networking_ansible.ansible_networking.'
            'AnsibleNetworking.vlan_access_port')
class TestDeletePortPostCommit(base.NetworkingAnsibleTestCase):
    def test_delete_port_postcommit_current(self,
                                            mock_vlan_access,
                                            mock_port_bound):
        self.mech.delete_port_postcommit(self.mock_port_context)
        mock_vlan_access.assert_called_once_with(
            'remove',
            self.mock_port_context.current,
            self.mock_port_context.network.current)

    def test_delete_port_postcommit_not_bound(self,
                                              mock_vlan_access,
                                              mock_port_bound):
        mock_port_bound.return_value = False
        self.mech.delete_port_postcommit(self.mock_port_context)
        mock_vlan_access.assert_not_called()


@mock.patch('networking_ansible.config.build_ansible_inventory', autospec=True)
class TestInit(base.NetworkingAnsibleTestCase):
    def test_intialize(self, mock_inventory):
        self.mech.initialize()
        mock_inventory.assert_called_once_with()


@mock.patch('networking_ansible.ml2.mech_driver.'
            'AnsibleMechanismDriver._is_port_bound')
@mock.patch('networking_ansible.ansible_networking.'
            'AnsibleNetworking.vlan_access_port')
@mock.patch('networking_ansible.ml2.mech_driver.provisioning_blocks',
            autospec=True)
class TestUpdatePortPostCommit(base.NetworkingAnsibleTestCase):
    def test_update_port_postcommit_current(self,
                                            mock_prov_blks,
                                            mock_vlan_access,
                                            mock_port_bound):
        self.mech.update_port_postcommit(self.mock_port_context)
        mock_prov_blks.provisioning_complete.assert_called_once()

    def test_update_port_postcommit_original(self,
                                             mock_prov_blks,
                                             mock_vlan_access,
                                             mock_port_bound):
        mock_port_bound.side_effect = [False, True]
        self.mech.update_port_postcommit(self.mock_port_context)
        mock_vlan_access.assert_called_once()

    def test_update_port_postcommit_not_bound(self,
                                              mock_prov_blks,
                                              mock_vlan_access,
                                              mock_port_bound):
        mock_port_bound.side_effect = [False, False]
        self.mech.update_port_postcommit(self.mock_port_context)
        mock_vlan_access.assert_not_called()
